/* mincsample.c                                                               */
/*                                                                            */
/* Generate samplings from MINC files                                         */
/*                                                                            */
/* Andrew Janke - rotor@cmr.uq.edu.au                                         */
/* Mark Griffin - mark.griffin@cmr.uq.edu.au                                  */
/* Center for Magnetic Resonance                                              */
/* The University of Queensland                                               */
/* http://www.cmr.uq.edu.au/~rotor                                            */
/*                                                                            */
/* Copyright (C) 2003 Andrew Janke and Mark Griffin                           */
/* This program is free software; you can redistribute it and/or              */
/* modify it under the terms of the GNU General Public License                */
/* as published by the Free Software Foundation; either version 2             */
/* of the License, or (at your option) any later version.                     */
/*                                                                            */
/* This program is distributed in the hope that it will be useful,            */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of             */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the              */
/* GNU General Public License for more details.                               */
/*                                                                            */
/* You should have received a copy of the GNU General Public License          */
/* along with this program; if not, write to the Free Software                */
/* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. */

#include <config.h>
#include <stdio.h>
#include <sys/time.h>
#include <unistd.h>
#include <math.h>
#include <string.h>
#include <ParseArgv.h>
#include <time_stamp.h>
#include <voxel_loop.h>
#include <gsl/gsl_rng.h>

#ifndef  FALSE
#define  FALSE  0
#endif

#ifndef  TRUE
#define  TRUE   1
#endif

/* typedefs */
typedef enum { SAMPLE_ALL, SAMPLE_RND } Sample_enum;
typedef enum { OUTPUT_ASCII, OUTPUT_DOUBLE } Output_enum;

typedef struct {
   Sample_enum sample_type;

   /* input parameters */
   int      masking;
   double   mask_val;
   int      mask_idx;

   /* sampling */
   gsl_rng *random;
   int      rand_samples;
   int      max_samples;

   /* output parameters */
   int      sample_mask;
   int      sample_mask_idx;

   Output_enum output_type;
   FILE    *outFP;
   } Loop_Data;

/* function prototypes */
void     count_points(void *caller_data, long num_voxels, int input_num_buffers,
                      int input_vector_length, double *input_data[],
                      int output_num_buffers, int output_vector_length,
                      double *output_data[], Loop_Info * loop_info);
void     get_points(void *caller_data, long num_voxels, int input_num_buffers,
                    int input_vector_length, double *input_data[], int output_num_buffers,
                    int output_vector_length, double *output_data[],
                    Loop_Info * loop_info);
void     write_data(FILE * fp, double value, Output_enum ot);
void     print_version_info(void);

/* Argument variables and table */
static int verbose = FALSE;
static int quiet = FALSE;
static int clobber = FALSE;
static int max_buffer = 4 * 1024;
static char *mask_fname = NULL;
static char *sample_fname = NULL;
static char *out_fname = NULL;
static int append_output = FALSE;
static Loop_Data md = {
   SAMPLE_ALL,
   FALSE, 1.0, 0,
   NULL, 0, 0,
   FALSE, 0,
   OUTPUT_ASCII, NULL
   };

static ArgvInfo argTable[] = {
   {NULL, ARGV_HELP, (char *)NULL, (char *)NULL,
    "General options:"},
   {"-version", ARGV_FUNC, (char *)print_version_info, (char *)NULL,
    "print version info and exit."},
   {"-verbose", ARGV_CONSTANT, (char *)TRUE, (char *)&verbose,
    "print out extra information."},
   {"-quiet", ARGV_CONSTANT, (char *)TRUE, (char *)&quiet,
    "be very quiet."},
   {"-clobber", ARGV_CONSTANT, (char *)TRUE, (char *)&clobber,
    "clobber existing files."},
   {"-max_buffer", ARGV_INT, (char *)1, (char *)&max_buffer,
    "maximum size of buffers (in kbytes)"},
   {"-mask", ARGV_STRING, (char *)1, (char *)&mask_fname,
    "select voxels within the specified mask"},
   {"-mask_val", ARGV_FLOAT, (char *)1, (char *)&(md.mask_val),
    "mask value to use"},

   {NULL, ARGV_HELP, (char *)NULL, (char *)NULL,
    "\nSampling Types:"},
   {"-all", ARGV_CONSTANT, (char *)SAMPLE_ALL, (char *)&md.sample_type,
    "sample all the data (Default)"},
   {"-random_samples", ARGV_INT, (char *)1, (char *)&md.rand_samples,
    "take # random samples from the input data"},

   {NULL, ARGV_HELP, (char *)NULL, (char *)NULL,
    "\nOutput Options:"},
   {"-sample", ARGV_STRING, (char *)1, (char *)&sample_fname,
    "Output a <mask.mnc> file of chosen points"},
   {"-outfile", ARGV_STRING, (char *)1, (char *)&out_fname,
    "<file> for output data (Default: stdout)"},
   {"-append", ARGV_CONSTANT, (char *)TRUE, (char *)&append_output,
    "append output data to existing file"},
   {"-ascii", ARGV_CONSTANT, (char *)OUTPUT_ASCII, (char *)&md.output_type,
    "Write out data as ascii strings (default)"},
   {"-double", ARGV_CONSTANT, (char *)OUTPUT_DOUBLE, (char *)&md.output_type,
    "Write out data as double precision floating-point values"},

   {NULL, ARGV_HELP, NULL, NULL, ""},
   {NULL, ARGV_END, NULL, NULL, NULL}
   };

int main(int argc, char *argv[])
{
   char   **infiles;
   char    *outfiles[1];
   int      n_infiles, n_outfiles;
   char    *arg_string;
   Loop_Options *loop_opts;
   struct timeval timer;
   int      i;

   /* Save time stamp and args */
   arg_string = time_stamp(argc, argv);

   /* get arguments */
   if(ParseArgv(&argc, argv, argTable, 0) || (argc < 2)){
      fprintf(stderr, "\nUsage: %s [options] <in1.mnc> <in2.mnc> ... \n", argv[0]);
      fprintf(stderr, "       %s -help\n\n", argv[0]);
      exit(EXIT_FAILURE);
      }

   /* check arguments */
   if(md.rand_samples != 0){
      md.sample_type = SAMPLE_RND;
      if(md.rand_samples < 0){
         fprintf(stderr, "%s: -rand_samples (%d) must be greater than 0\n\n",
                 argv[0], md.rand_samples);
         exit(EXIT_FAILURE);
         }
      }

   /* get infile names */
   n_infiles = argc - 1;
   infiles = (char **)malloc(sizeof(char *) * (n_infiles + 1));   /* + 1 for mask */
   for(i = 0; i < n_infiles; i++){
      infiles[i] = argv[i + 1];
      }
   if(mask_fname != NULL){
      infiles[n_infiles] = mask_fname;
      md.masking = TRUE;
      md.mask_idx = n_infiles;
      n_infiles++;
      }

   /* check for the infile(s) */
   for(i = 0; i < n_infiles; i++){
      if(access(infiles[i], F_OK) != 0){
         fprintf(stderr, "%s: Couldn't find input file: %s\n\n", argv[0], infiles[i]);
         exit(EXIT_FAILURE);
         }
      }

   /* set up and check for voxel_loop outfiles */
   if(sample_fname != NULL){
      if(access(sample_fname, F_OK) == 0 && !clobber){
         fprintf(stderr, "%s: %s exists, use -clobber to overwrite\n\n", argv[0],
                 sample_fname);
         exit(EXIT_FAILURE);
         }
      md.sample_mask = TRUE;
      md.sample_mask_idx = 0;
      n_outfiles = 1;
      outfiles[0] = sample_fname;
      }
   else {
      n_outfiles = 0;
      }

   /* set up data outfile */
   if(out_fname == NULL || strcmp(out_fname, "-") == 0){
      md.outFP = stdout;
      }
   else {
      if(!append_output && access(out_fname, F_OK) == 0 && !clobber){
         fprintf(stderr, "%s: %s exists, use -clobber to overwrite\n\n", argv[0],
                 out_fname);
         exit(EXIT_FAILURE);
         }

      if((md.outFP = fopen(out_fname, (append_output) ? "a" : "w")) == NULL){
         fprintf(stderr, "%s:  problems opening %s\n", argv[0], out_fname);
         exit(EXIT_FAILURE);
         }
      }

   /* set up voxel loop options */
   loop_opts = create_loop_options();
   set_loop_verbose(loop_opts, FALSE);
   set_loop_clobber(loop_opts, clobber);
   set_loop_buffer_size(loop_opts, (long)1024 * max_buffer);

   /* get and check the max number of points if required */
   if(md.sample_type == SAMPLE_RND && md.max_samples == 0){
      voxel_loop(1, (md.masking) ? &mask_fname : infiles, 0, NULL, NULL,
                 loop_opts, count_points, (void *)&md);
      if(verbose){
         fprintf(stderr, "| Got max # of points: %d\n", md.max_samples);
         }
      }
   if(md.max_samples < 0){
      fprintf(stderr, "%s:  -max_samples (%d) must be positive\n\n", argv[0],
              md.max_samples);
      exit(EXIT_FAILURE);
      }

   /* set up random sampling if required */
   if(md.sample_type == SAMPLE_RND){
      void    *tmp = NULL;             /* for gettimeofday */

      if(md.rand_samples > md.max_samples){
         fprintf(stderr, "%s: -rand_samples (%d) must be less than -max_samples (%d)\n\n",
                 argv[0], md.rand_samples, md.max_samples);
         exit(EXIT_FAILURE);
         }

      /* initialise random number generator */
      md.random = gsl_rng_alloc(gsl_rng_mt19937);
      gettimeofday(&timer, tmp);
      gsl_rng_set(md.random, timer.tv_usec);
      }

   /* do the sampling */
   voxel_loop(n_infiles, infiles, n_outfiles, outfiles, arg_string,
              loop_opts, get_points, (void *)&md);

   /* tidy up */
   fclose(md.outFP);
   free_loop_options(loop_opts);
   if(md.sample_type == SAMPLE_RND){
      gsl_rng_free(md.random);
      }

   return (EXIT_SUCCESS);
   }

/* get points from file(s), write out to an input FP */
void get_points(void *caller_data, long num_voxels, int input_num_buffers,
                int input_vector_length, double *input_data[], int output_num_buffers,
                int output_vector_length, double *output_data[], Loop_Info * loop_info)
{
   Loop_Data *md = (Loop_Data *) caller_data;
   int      i, ivox;
   int      n_infiles;
   double   mask_value;

   /* shut the compiler up */
   (void)output_num_buffers;
   (void)output_vector_length;
   (void)loop_info;

   n_infiles = (md->masking) ? input_num_buffers - 1 : input_num_buffers;

   /* for each voxel */
   for(ivox = 0; ivox < num_voxels * input_vector_length; ivox++){

      /* nasty way that works for masking or not */
      mask_value = 0;
      if(!md->masking ||
         (md->masking && fabs(input_data[md->mask_idx][ivox] - md->mask_val) < 0.5)){

         switch (md->sample_type){
         case SAMPLE_ALL:
            for(i = 0; i < n_infiles; i++){
               write_data(md->outFP, input_data[i][ivox], md->output_type);
               }
            mask_value = 1.0;
            break;

         case SAMPLE_RND:
            /* if this voxel 'qualifies', write out data */
            if(gsl_rng_uniform(md->random) < ((double)md->rand_samples / md->max_samples)){
               md->rand_samples--;
               for(i = 0; i < n_infiles; i++){
                  write_data(md->outFP, input_data[i][ivox], md->output_type);
                  }
               mask_value = 1.0;
               }

            md->max_samples--;
            break;

         default:
            fprintf(stderr, "ERROR - Sample type is undefined (%d)\n", md->sample_type);
            exit(EXIT_FAILURE);
            }
         }

      /* output sampling mask */
      if(md->sample_mask){
         output_data[md->sample_mask_idx][ivox] = mask_value;
         }
      }

   }

/* count points that are within a mask */
void count_points(void *caller_data, long num_voxels, int input_num_buffers,
                  int input_vector_length, double *input_data[],
                  int output_num_buffers, int output_vector_length,
                  double *output_data[], Loop_Info * loop_info)
{
   Loop_Data *md = (Loop_Data *) caller_data;
   int      ivox;

   /* shut the compiler up */
   (void)input_num_buffers;
   (void)output_num_buffers;
   (void)output_vector_length;
   (void)output_data;
   (void)loop_info;

   for(ivox = 0; ivox < num_voxels * input_vector_length; ivox++){
      if(md->masking){
         if(fabs(input_data[0][ivox] - md->mask_val) < 0.5){
            md->max_samples++;
            }
         }
      else {
         md->max_samples++;
         }
      }
   }

void write_data(FILE * fp, double value, Output_enum ot)
{
   switch (ot){
   case OUTPUT_ASCII:
      fprintf(fp, "%.20g\n", value);
      break;

   case OUTPUT_DOUBLE:
      fwrite(&(value), sizeof(double), 1, fp);
      break;

   default:
      fprintf(stderr, "ERROR - Output type is undefined (%d)\n", ot);
      exit(EXIT_FAILURE);
      }
   }

void print_version_info(void)
{
   fprintf(stdout, "\n");
   fprintf(stdout, "%s version %s\n", PACKAGE, VERSION);
   fprintf(stdout, "Comments to %s\n", PACKAGE_BUGREPORT);
   fprintf(stdout, "\n");
   exit(EXIT_SUCCESS);
   }
